#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
:mod:`analyse_tris` module
:author: FIL - Faculté des Sciences et Technologies - Univ. Lille <http://portail.fil.univ-lille1.fr>_
:date: janvier 2017
:dernières révisions: février 2018, février 2019

Analyse empirique des tris

"""
from random import shuffle
from typing import Callable
from compare import compare
from ap_decorators import count
from tris import *

################################################
#  ANALYSE EMPIRIQUE DES TRIS                  #
################################################

# ajout d'un compteur à la fonction compare
compare = count(compare)

def analyser_tri(tri: Callable[[list[T], Callable[[T, T], int]], NoneType],
                 nbre_essais: int,
                 taille: int) -> float:
    """
    renvoie: le nombre moyen de comparaisons effectuées par l'algo tri
         pour trier des listes de taille t, la moyenne étant calculée
         sur n listes aléatoires.
    précondition: n > 0, t >= 0, la fonc
    """
    res = 0
    for i in range(nbre_essais):
        compare.counter = 0
        l = [k for k in range(taille)]
        shuffle(l)
        tri(l, compare)
        res += compare.counter
    return res / nbre_essais

def analyser_sort(sort: Callable[list[T], list], nb_essai:int, taille:int)->float:
    res = 0
    for i in range(nb_essai):
        compare.counter = 0
        l = [k for k in range(taille)]
        shuffle(l)
        sort(l)
        res += compare.counter
    return res / nb_essai
    


if (__name__ == '__main__'):
    from matplotlib import pyplot as plt
    
    # Calcul de nombres moyens de comparaison pour des listes
    # de tailles comprises entre 0 et TAILLE_MAX
    NB_ESSAIS = 50
    TAILLE_MAX = 100
    c_select = [0.0] * (TAILLE_MAX + 1)
    c_insert = [0.0] * (TAILLE_MAX + 1)
    c_sort = [0,0] * (TAILLE_MAX + 1)
    
    for t in range(TAILLE_MAX + 1):
        c_select[t] = analyser_tri(tri_select, 1, t)
        # inutile de moyenner pour le tri par sélection
        c_insert[t] = analyser_tri(tri_insert, NB_ESSAIS, t)
        c_sort[t] = analyser_sort(sorted, NB_ESSAIS, t)

    # Sauvegarde des données calculées dans un fichier au format CSV
    prem_ligne = 'taille;"tri séléction";"tri insertion";"tri avec methode sort"\n'
    ligne = '{:3d};{:8.2f};{:8.2f};{:8.2f}\n'
    with open('analyse_tris.csv', 'wt', encoding='utf8') as sortie:
        sortie.write(prem_ligne)
        for t in range(TAILLE_MAX + 1):
            sortie.write(ligne.format(t,
                                      c_select[t],
                                      c_insert[t],
                                      c_sort[t]))

    # Représentation graphique
    plt.plot(list(range(TAILLE_MAX + 1)), c_select, 'b.', label='Tri sélection')
    plt.plot(list(range(TAILLE_MAX + 1)), c_insert, 'r.', label='Tri insertion')
    plt.plot(list(range(TAILLE_MAX + 1)), c_sort, 'black', label='Tri avec sort')
    plt.title('Tris : nbre de comparaisons')
    plt.legend()
    plt.xlabel('n = taille des listes')
    plt.ylabel('c(n) = nbre de comparaisons')
    plt.savefig('tris_nbcomp.png')
    plt.show()


